from . import *
from time import time


def _test():
    print("Test: phi computing (100 digits)")
    s = time()
    print(phi())
    print("{} seconds".format(time() - s))
    print("Test: phi computing (1000 digits)")
    s = time()
    print(phi(1000))
    print("{} seconds".format(time() - s))
    print("Test: pi computing (100 digits)")
    s = time()
    print(pi())
    print("{} seconds".format(time() - s))
    print("Test: pi computing (1000 digits)")
    s = time()
    print(pi(1000))
    print("{} seconds".format(time() - s))
    print("Test: Square root (100 digits)")
    s = time()
    print(sqrt(Real(2)))
    print("{} seconds".format(time() - s))
    print("Test: Square root (1000 digits)")
    s = time()
    print(sqrt(Real(2, precision=1000)))
    print("{} seconds".format(time() - s))


if __name__ == "__main__":
    _test()



"""
Provides the Real object for exact arithmetic
"""
from math import gcd
from functools import wraps


def simplifier(func):
    @wraps(func)
    def infunc(self, *args, **kwargs):
        ret = func(self, *args, **kwargs)
        if not self.simplify:
            return ret
        g = gcd(self.numerator, self.denominator)
        self.numerator //= g
        self.denominator //= g
        return ret

    return infunc


class Real:
    precision = 100  # Default precision
    simplify = True  # To or not to simplify the fraction after an operation (Note: without simplification it will be much faster)

    def __init__(self, *args, simplify=True, precision=100):
        """
        Provides a lossless real number
        Real(value)
        Converts the given value into Real object, supports integer, string, another Real and float
        For example:
        >>> Real(1)
        1
        >>> Real("2")
        2
        >>> Real(0.5)
        0.5
        >>> Real(0.1)
        0.1000000000000000055511151231257827021181583404541015625
        Real(numerator,denominator)
        Converts the fraction numerator/denominator into Real object, it is exact.
        >>> Real(1,10)
        0.1
        >>> Real(Real(1))
        1

        the precision and simplify arguments controls precision and simplification
        """
        self.precision = precision
        self.simplify = simplify
        if len(args) == 2:
            self.numerator, self.denominator = args
        elif len(args) == 1:
            if isinstance(args[0], int):
                self.numerator, self.denominator = args[0], 1
            elif isinstance(args[0], str):
                pos = args[0].find(".")
                if pos == -1:
                    self.numerator, self.denominator = int(args[0]), 1
                else:
                    self.numerator, self.denominator = int(
                        args[0].replace(".", "")
                    ), 10 ** len(args[0][pos + 1 :])
            elif isinstance(args[0], float):
                s = "%.100f" % args[0]
                pos = s.find(".")
                if pos == -1:
                    self.numerator, self.denominator = int(s), 1
                else:
                    a, b = s.split(".")
                    self.numerator, self.denominator = int(
                        a + b.rstrip("0")
                    ), 10 ** len(b.rstrip("0"))
            elif isinstance(
                args[0], Real
            ):  # In this case, the precision argument and simplify arguments are ignored
                self.precision = args[0].precision
                self.numerator = args[0].numerator
                self.denominator = args[0].denominator
                self.simplify = args[0].simplify
            else:
                raise TypeError("Only supports int, str, Real and float")
        else:
            raise ValueError("Real class only accepts one or two positional arguments")
        if self.simplify:
            g = gcd(self.numerator, self.denominator)
            self.numerator //= g
            self.denominator //= g

    @simplifier
    def tofraction(self):
        """
        Convert self to a fraction string
        >>> a=Real(0.1)
        >>> a.tofraction()
        '3602879701896397/36028797018963968'
        >>> b=Real(0.1,simplify=False)
        >>> b.tofraction() # If you don't want the number to be simplified, the fraction remains unsimplified, too.
        1000000000000000055511151231257827021181583404541015625/10000000000000000000000000000000000000000000000000000000
        :return: The fraction as a string, "a/b" if it is not an integer, or "a" if it is an integer
        """
        if self.denominator == 1:
            return str(self.numerator)
        if self.denominator == -1:
            return "-" + str(self.numerator)
        return str(abs(self.numerator) * self.sign()) + "/" + str(abs(self.denominator))

    @simplifier
    def tostring(self):
        """
        Convert self to a decimal string, this operation removes the trailing zeros
        >>> a=Real(0.1) # Python's built-in float is inaccurate
        >>> a.tostring()
        '0.1000000000000000055511151231257827021181583404541015625'
        >>> a.precision=10
        >>> a.tostring()
        '0.1'
        >>> b=Real("0.1") # Exact
        >>> b.tostring()
        '0.1'
        >>> c=Real(1)/Real(7)
        >>> c.tostring()
        '0.1428571428571428571428571428571428571428571428571428571428571428571428571428571428571428571428571428'
        >>> c.precision=10
        >>> c.tostring() # Truncated output string
        '0.1428571428'
        >>> c.precision=100
        >>> c.tostring() # Precision does not lose after truncation
        '0.1428571428571428571428571428571428571428571428571428571428571428571428571428571428571428571428571428'
        :return: The decimal string
        """
        self.numerator *= self.sign()
        self.denominator = abs(self.denominator)
        intpart, floatpart = (
            self.sign() * (abs(self.numerator) // self.denominator),
            abs(self.numerator) % self.denominator,
        )
        result = str(intpart)
        result_floatpart = ""
        if self.precision <= 0:
            return result
        for i in range(self.precision):  # Multiply 10 repeatedly
            floatpart *= 10
            result_floatpart += str(floatpart // self.denominator)
            floatpart %= self.denominator
        result_floatpart = result_floatpart.rstrip("0")
        if not result_floatpart:
            return result
        return result + "." + result_floatpart

    def __truediv__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return Real(
            self.numerator * b.denominator,
            self.denominator * b.numerator,
            simplify=self.simplify or b.simplify,
            precision=max(self.precision, b.precision),
        )

    def __floordiv__(self, b):  # Warning: This returns a regular Python integer
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return (self.numerator * b.denominator) // (self.denominator * b.numerator)

    def __trunc__(self):
        return self.numerator // self.denominator

    def __float__(self):
        return self.numerator / self.denominator

    def sign(self):
        """
        Get the sign of self
        :return: -1 if self is negative, 0 if self is zero, and 1 if self is positive
        """
        if self.numerator * self.denominator == 0:
            return 0
        return (
            self.numerator * self.denominator // abs(self.numerator * self.denominator)
        )

    def __mul__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return Real(
            self.numerator * b.numerator,
            self.denominator * b.denominator,
            simplify=self.simplify or b.simplify,
            precision=max(self.precision, b.precision),
        )

    def __add__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return Real(
            self.numerator * b.denominator + self.denominator * b.numerator,
            self.denominator * b.denominator,
            simplify=self.simplify or b.simplify,
            precision=max(self.precision, b.precision),
        )

    def __sub__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return Real(
            self.numerator * b.denominator - self.denominator * b.numerator,
            self.denominator * b.denominator,
            simplify=self.simplify or b.simplify,
            precision=max(self.precision, b.precision),
        )

    def __neg__(self):
        return Real(
            -self.numerator,
            self.denominator,
            simplify=self.simplify,
            precision=self.precision,
        )

    def __abs__(self):
        return Real(
            abs(self.numerator),
            abs(self.denominator),
            simplify=self.simplify,
            precision=self.precision,
        )

    def __gt__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return self.numerator * b.denominator > self.denominator * b.numerator

    def __eq__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return self.numerator * b.denominator == self.denominator * b.numerator

    def __ne__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return self.numerator * b.denominator != self.denominator * b.numerator

    def __lt__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return self.numerator * b.denominator < self.denominator * b.numerator

    def __ge__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return self.numerator * b.denominator >= self.denominator * b.numerator

    def __le__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return self.numerator * b.denominator <= self.denominator * b.numerator

    def __iadd__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        x = self.__add__(b)
        self.numerator, self.denominator = x.numerator, x.denominator
        return self

    def __isub__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        x = self.__sub__(b)
        self.numerator, self.denominator = x.numerator, x.denominator
        return self

    def __imul__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        x = self.__mul__(b)
        self.numerator, self.denominator = x.numerator, x.denominator
        return self

    def __itruediv__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        x = self.__truediv__(b)
        self.numerator, self.denominator = x.numerator, x.denominator
        return self

    def __ifloordiv__(
        self, b
    ):  # Warning: This results in a Real object with the same value as the corresponding integer
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        self.numerator, self.denominator = self.__floordiv__(b), 1
        return self

    def __deepcopy__(self, *_, **__):
        return Real(
            self.numerator,
            self.denominator,
            precision=self.precision,
            simplify=self.simplify,
        )

    def __hash__(self):
        return hash((self.numerator, self.denominator))

    def __pow__(self, b, *_, **__):
        from .algorithms import pow

        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        return pow(self, b)

    def __ipow__(self, b, *_, **__):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        x = self.__pow__(b)
        self.numerator, self.denominator = x.numerator, x.denominator
        return self

    def __mod__(self, b):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        if not b:  # Any number % 0 is 0
            return Real(0, simplify=self.simplify, precision=self.precision)
        return Real(
            (self.numerator * b.denominator) % (self.denominator * b.numerator),
            self.denominator * b.numerator,
            simplify=self.simplify,
            precision=self.precision,
        )

    def __imod__(self, b, *_, **__):
        if not isinstance(b, Real):
            b = Real(b, simplify=self.simplify, precision=self.precision)
        x = self.__mod__(b)
        self.numerator, self.denominator = x.numerator, x.denominator
        return self

    __int__ = __trunc__
    __bool__ = lambda _: _.numerator != 0
    __repr__ = lambda *_: _[0].tostring(*_[1:])
    __str__ = lambda *_: _[0].tostring(*_[1:])


__all__ = ["Real"]



"""
precreal is a Python package, allowing you to compute floating point numbers without precision loss.

It is in pure Python and doesn't even use ctypes, so it is cross-platform and doesn't require a C/C++ library.

Works in Python 3, because in Python 3, integers are bignums, doesn't and never will work in Python 2.
"""
import sys as __sys

if hasattr(__sys, "set_int_max_str_digits"):
    __sys.set_int_max_str_digits(0)
from .precreal import *
from .algorithms import *



from setuptools import setup
setup(
    name="precreal",
    version="1.0.0",
    description="A package providing a fraction-based Real class that allows you to do exact real arithmetic, along with some algorithms",
    author="None1",
    packages=["precreal"],
    keywords=["arithmetic", "calculating", "real number"],
    python_requires=">=3",
    url="https://github.com/none-None1/precreal",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    entrypoints={"console_scripts": ["precreal_test=precreal.test:_test"]},
)



"""
Some algorithms for the Real type in precreal
"""
from .precreal import Real
from copy import deepcopy


def intpow(x, y):
    """
    Computes x**y
    :param x: Real
    :param y: integer
    :return: Real
    """
    a = deepcopy(x, {})
    result = Real(1, precision=x.precision, simplify=x.simplify)
    while y:
        if y % 2:
            result *= a
        y //= 2
        a *= a
    return result


def root(x, y):
    """
    Computes x**(1/y) (up to x.precision)
    :param x: Real
    :param y: integer
    :return: x**(1/y)
    """
    if y % 2 == 0 and x.sign() == -1:
        raise ValueError("Result not a real number")
    z = Real(1, 10**x.precision, precision=x.precision, simplify=x.simplify)
    l, r = Real(0, precision=x.precision, simplify=x.simplify), max(
        x, Real(1, precision=x.precision, simplify=x.simplify)
    )
    while abs(l - r) >= z:
        mid = (l + r) / 2
        if intpow(mid, y) > x:
            r = mid
        else:
            l = mid
    return l


def pow(x, y):
    """
    Computes x**y (up to x.precision), if the exponent is an integer, it is recommended to use intpow() instead
    >>> pow(Real(2),Real(1,2))
    1.4142135623730950488016887242096980785696718753769480731766797379907324784621070388503875343276415727
    >>> pow(Real(2),Real(3))
    8
    :param x: Real
    :param y: Real
    :return: Real
    """
    if not isinstance(y, Real):
        y = Real(y)
    if not isinstance(x, Real):
        x = Real(x)
    result = root(intpow(x, y.numerator), y.denominator)
    if y.sign == -1:
        result = Real(
            result.denominator,
            result.numerator,
            simplify=result.simplify,
            precision=result.precision,
        )
    return result


sqrt = lambda x: root(x, 2)


def phi(prec=100):
    """
    Computes phi
    >>> phi()
    1.6180339887498948482045868343656381177203091798057628621354486227052604628189024497072072041893911374
    >>> phi(2)
    1.61
    :param prec: digits
    :return: a Real object with precision prec, represents phi
    """
    return (Real(1, precision=prec) + sqrt(Real(5, precision=prec))) / Real(
        2, precision=prec
    )


def pi(prec=100):
    """
    Computes pi
    >>> pi()
    3.14159265358979323846264338327950288419716939937510582097494459230781640628620899862803482534211706798214808651
    >>> pi(2)
    3.14
    :param prec: digits
    :return: a Real object with precision prec, represents pi
    """
    prec += 10
    r = Real(0, precision=prec)
    for i in range(prec):
        r += intpow(Real(1, 5, precision=prec), i * 4 + 1) / (i * 4 + 1) - intpow(
            Real(1, 5, precision=prec), i * 4 + 3
        ) / (i * 4 + 3)
    s = Real(0, precision=prec)
    for i in range(prec):
        s += intpow(Real(1, 239, precision=prec), i * 4 + 1) / (i * 4 + 1) - intpow(
            Real(1, 239, precision=prec), i * 4 + 3
        ) / (i * 4 + 3)
    a = r * 16 - s * 4
    a.precision = prec
    return a


def rand(prec=100, use_secrets=True):
    """
    Return a random Real number in [0,1)
    >>> rand()
    0.6430997410264154172521480037843260492523013140407526344628086475835143370634611004596928411125970842
    >>> rand(2)
    0.41
    :param prec: digits
    :param use_secrets: Whether to use the "secrets" module in Python if available, if this is set to False or if the "secrets" module is unavailable, it uses the "random" module
    :return: a random number
    """
    if use_secrets:
        try:
            import secrets

            return Real(secrets.randbelow(10**prec), 10**prec, precision=prec)
        except:
            from random import randint

            return Real(randint(0, 10**prec - 1), 10**prec, precision=prec)
    else:
        from random import randint

        return Real(randint(0, 10**prec - 1), 10**prec, precision=prec)


__all__ = ["pow", "sqrt", "phi", "pi", "rand", "intpow"]
